""" fURI File access through URIs. """


__author__  = "amancevice"
__email__   = "smallweirdnum@gmail.com"
__version__ = "0.6.7"


from .furimap import chain
from .utils import add_handler
from .utils import download
from .utils import exists
from .utils import map
from .utils import open
from .utils import walk

try:
    from . import aws
except ImportError:
    pass

try:
    from . import sftp
except ImportError:
    pass



""" AWS backed File and FileMap implementations. """

import collections
import os
import re
import boto3
import botocore
from urlparse import urlparse
from . import furifile
from . import utils


class S3File(furifile.RemoteFile):
    """ S3-backed file implementation.

        Ex. s3://bucket/path/to/key """

    @property
    def bucket(self):
        """ Remote bucket. """
        try:
            return self._bucket
        except AttributeError:
            self._bucket = self.connection.Bucket(self.uri.netloc)
            return self._bucket

    @property
    def key(self):
        """ Remote key. """
        try:
            return self._key
        except AttributeError:
            self._key = self.connection.Object(
                self.uri.netloc, self.uri.path.lstrip('/'))
            return self._key

    def connect(self, **connectkw):
        if connectkw:
            self.__connect__ = connectkw

        # Support boto-style access/secret keys
        if 'access_key' in self.__connect__:
            self.__connect__['aws_access_key_id'] = self.__connect__['access_key']
            del self.__connect__['access_key']
        if 'secret_key' in self.__connect__:
            self.__connect__['aws_secret_access_key'] = self.__connect__['secret_key']
            del self.__connect__['secret_key']

        # Connect to AWS
        self._connection = boto3.resource('s3', **self.__connect__)
        return self._connection

    def download(self, target):
        """ Download remote file to local target URI. """
        self.key.download_file(str(target))
        return target

    def exists(self):
        """ Test file existence. """
        try:
            self.key.load()
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'] == "404":
                return False
            raise err
        return True

    def write(self, stream):
        """ Write stream to file. """
        try:
            return self.key.put(Body=stream)
        except AttributeError:
            return self.key.put(Body=stream.read())

    def _stream_impl(self):
        """ Implementation of stream(). """
        return self.key.get().get('Body')

    def _walk_impl(self, **kwargs):
        """ Implementation of walk(). """
        root = str(re.sub('^/', '', self.uri.path))
        tree = { root : { 'dirnames' : set(), 'filenames' : set() } }
        for key in self.bucket.objects.filter(Prefix=self.key.key):
            rel, filename = map(str, os.path.split(re.split("^%s" % root, key.key)[-1]))
            history = root
            for subdir in rel.split('/'):
                if subdir:
                    tree[history]['dirnames'].add(subdir)
                history = os.path.normpath(os.path.join(history, subdir)) + '/'
                if history not in tree:
                    tree[history] = { 'dirnames' : set(), 'filenames' : set() }
            if filename:
                tree[history]['filenames'].add(filename)

        for dirpath in sorted(tree.keys()):
            dirdata   = tree[dirpath]
            dirnames  = sorted(dirdata['dirnames'])
            filenames = sorted(dirdata['filenames'])
            dirpath = "s3://%s/%s" % (self.bucket.name, dirpath)
            yield dirpath, dirnames, filenames


class DynamoMap(collections.Iterable):
    """ DynamoDB-backed mappings. """
    def __init__(self, uri, **connectkw):
        self.uri = urlparse(uri)
        self.connection = boto3.resource('dynamodb', **connectkw)
        self.table = self.connection.Table(self.uri.netloc)

    def __str__(self):
        return self.uri.geturl()

    def __repr__(self):
        return "<%s: %s>" % (type(self).__name__, self.uri.geturl())

    def __getitem__(self, key):
        try:
            ikey = { self.table.key_schema[0]['AttributeName'] : key }
            return self.__read(self.table.get_item(Key=ikey))['Item']
        except KeyError:
            raise KeyError(key)

    def __iter__(self):
        response = self.table.scan()
        items = response.get('Items') or []
        for item in items:
            yield item
        while 'LastEvaluatedKey' in response:
            response = self.table.scan(ExclusiveStartKey=response['LastEvaluatedKey'])
            items = response.get('Items') or []
            for item in items:
                yield item

    def __len__(self):
        return self.table.item_count

    def __setitem__(self, key, value):
        origin = self[key].get('Item', key).items()
        update = value.items()
        return self.table.put_item(Item=dict(origin + update))

    def __delitem__(self, key):
        self.table.delete_item(Key=key)

    def __read(self, results=None):
        """ Read Dynamo result and throw ValueError on non-200 response code. """
        results = results or self.table.scan()
        response = results.get('ResponseMetadata') or {}
        httpcode = response.get('HTTPStatusCode')
        if httpcode != 200:
            raise ValueError(response)
        return results


utils.add_handler('s3', S3File, 'file')
utils.add_handler('dynamodb', DynamoMap, 'map')



""" fURI Mappings. """

import collections
import json
import logging
import os
import yaml
from . import utils


class FileMap(collections.Mapping):
    """ Base configuration for file objects. """
    __dispatch__ = {
        '.json' : json.loads,
        '.yaml' : yaml.load,
        '.yml'  : yaml.load}

    def __init__(self, uri, **kwargs):
        self.source = utils.open(uri, **kwargs)

    def __str__(self):
        return str(self.source)

    def __repr__(self):
        return "<%sMap: %s>" % (type(self.source).__name__, self.source)

    def __getitem__(self, key):
        try:
            return self.__read()[key]
        except Exception as err:
            raise KeyError(err.message)

    def __iter__(self):
        try:
            return iter(self.__read())
        except Exception as err:
            raise StopIteration(err.message)

    def __len__(self):
        try:
            return len(self.__read())
        except Exception as err:
            raise ValueError(err.message)

    def __read(self):
        """ Read contents and parse from __dispatch__. """
        ext = os.path.splitext(str(self))[-1]
        func = self.__dispatch__.get(ext, lambda x: x)
        return func(self.source.read())


class ChainedMap(collections.Mapping):
    """ Chained AWS or locally backed mappings. """
    def __init__(self, *mappings):
        self.mappings = mappings

    def __getitem__(self, key):
        return self.__read('__getitem__', key)

    def __iter__(self):
        return self.__read('__iter__')

    def __len__(self):
        return self.__read('__len__')

    def __read(self, func, *args, **kwargs):
        """ Try to access mappings in chain, return the first success.

            Arguments:
                func   (str):    Name of self-function to run against each cfg
                args   (tuple):  Tuple of arguments to func
                kwargs (dict):   Dictionary of arguments to func

            Returns:
                The return value of 'func(*args, **kwargs)' for the first
                cfg that does not throw an Exception. """
        for cfg in self.mappings:
            try:
                return getattr(cfg, func)(*args, **kwargs)
            except (KeyError, ValueError, StopIteration) as err:
                logging.warn("%s :: %r", cfg, err)
        raise err


def chain(*mappings):
    """ Chain mappings together. """
    return ChainedMap(*mappings)


utils.add_handler('' , FileMap, 'map')
utils.add_handler('file', FileMap, 'map')
utils.add_handler('s3', FileMap, 'map')



import os
import re
from setuptools import setup

NAME    = "furi"
AUTHOR  = "amancevice"
EMAIL   = "smallweirdnum@gmail.com"
DESC    = "fURI File access through URIs."

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2.7",
    "Topic :: Utilities" ]
TEST_REQUIRES  = ["boto", "nose", "mock", "moto"]
AWS_REQUIRES   = ["boto3>=1.2.3"]
SFTP_REQUIRES  = ["pysftp>=0.2.8"]
REQUIRES       = ["PyYAML>=3.11.0"]
EXTRAS_REQUIRE = {
    'aws'  : AWS_REQUIRES,
    'sftp' : SFTP_REQUIRES,
    'test' : TEST_REQUIRES,
    'all'  : AWS_REQUIRES + SFTP_REQUIRES + REQUIRES }

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def version():
    search = r"^__version__ *= *['\"]([0-9.]+)['\"]"
    initpy = read("./%s/__init__.py" % NAME)
    return re.search(search, initpy, re.MULTILINE).group(1)

setup(
    name                 = NAME,
    version              = version(),
    author               = AUTHOR,
    author_email         = EMAIL,
    packages             = [ NAME ],
    package_data         = { "%s" % NAME : ['README.md'] },
    include_package_data = True,
    url                  = 'http://www.smallweirdnumber.com',
    description          = DESC,
    long_description     = read('README.md'),
    classifiers          = CLASSIFIERS,
    install_requires     = REQUIRES,
    test_requires        = TEST_REQUIRES,
    extras_require       = EXTRAS_REQUIRE,
    test_suite           = "nose.collector" )



""" fURI Files. """

import collections
import os
import re
from urlparse import urlparse


MODES = 'r', 'rb', 'r+', 'rb+', 'w', 'wb', 'w+', 'wb+', 'a', 'ab', 'a+', 'ab+'


class File(collections.Iterable):
    """ Local File implentation.

        Exs. file:///abs/path/to/file.ext
             /abs/path/to/file.ext """

    modes = set(MODES)

    def __init__(self, uri, mode='r'):
        if mode not in self.modes and not set(mode).issubset(self.modes):
            raise ValueError("Cannot open %s in %s-mode" % (type(self).__name__, mode))
        self.uri  = urlparse(uri)
        self.path = self.uri.path
        self.mode = mode
        self.workdir, self.filename = os.path.split(self.path)

    def __str__(self):
        return self.uri.geturl()

    def __repr__(self):
        return "<%s: %s>" % (type(self).__name__, str(self))

    def __iter__(self):
        return self.stream()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """ Close stream. """
        if self.exists():
            return self.stream().close()

    def matches(self, pattern):
        """ Filename matches pattern.

            Arguments:
                pattern (regex or str):  Regular Expression to match.

            Returns:
                RegEx match object. """
        return re.compile(pattern).match(self.filename)

    def exists(self):
        """ Test file existence. """
        return os.path.exists(self.path)

    def read(self, *size):
        """ Read file stream. """
        return self.stream().read(*size)

    def write(self, stream):
        """ Write stream to file. """
        if not os.path.exists(self.workdir) and ('w' in self.mode or 'a' in self.mode):
            os.makedirs(self.workdir)
        try:
            return self.stream().write(stream)
        except TypeError:
            return self.stream().write(stream.read())

    def stream(self):
        """ Get file contents as stream. """
        try:
            self._stream.seek(0)
            return self._stream
        except AttributeError:
            if not self.exists() and 'w' not in self.mode:
                raise ValueError("%s does not exist" % self.uri.geturl())
            self._stream = self._stream_impl()
            return self._stream

    def walk(self, **kwargs):
        """ Walk the contents of a directory. """
        return self._walk_impl(**kwargs)

    def _walk_impl(self, **kwargs):
        """ Implementation of walk(). """
        return os.walk(self.path, **kwargs)

    def _stream_impl(self):
        """ Implementation of stream(). """
        return open(self.path, self.mode)


class RemoteFile(File):
    """ Remote file implentation. """

    def __init__(self, uri, mode='r', **connectkw):
        super(RemoteFile, self).__init__(uri, mode=mode)
        self.__connect__ = connectkw

    @property
    def connection(self):
        """ Remote connection. """
        try:
            return self._connection
        except AttributeError:
            self._connection = self.connect()
            return self._connection

    def connect(self, **kwargs):
        """ Connect to remote. """
        raise NotImplementedError

    def download(self, target):
        """ Download remote file to local target URI. """
        raise NotImplementedError

    def exists(self):
        """ Test file existence. """
        raise NotImplementedError

    def write(self, stream):
        """ Write stream to file. """
        raise NotImplementedError

    def _walk_impl(self, **kwargs):
        """ Implementation of walk(). """
        raise NotImplementedError

    def _stream_impl(self):
        """ Implementation of stream(). """
        raise NotImplementedError
